/*
 * Copyright (c) 2004 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#ifndef UTIL_H
# define UTIL_H

# include <sys/types.h>

# include <CoreFoundation/CoreFoundation.h>

extern int parseProperty(const char *, CFStringRef*, CFTypeRef*);
extern CFDictionaryRef ReadMetadata(const char*);
extern int WriteMetadata(const char *, CFDictionaryRef);
//extern uint32_t ChecksumData(CFDictionaryRef);
extern uint32_t ChecksumMetadata(const char *);
extern int UpdateChecksum(const char *, uint32_t);
extern int VerifyChecksum(const char *);
extern uint32_t GetChecksum(const char*);
extern uint32_t GetMetadataSize(const char*);
extern int IsAppleLabel(const char*);
extern uint32_t GetBlockSize(const char*);
extern uint64_t GetDiskSize(const char*);
extern int InitialMetadata(const char *, CFDictionaryRef, uint32_t);

extern int gDebug, gVerbose;

#endif /* UTIL_H */
